[![crates.io](https://img.shields.io/crates/v/calloop.svg)](https://crates.io/crates/calloop)
[![docs.rs](https://docs.rs/calloop/badge.svg)](https://docs.rs/calloop)
{{badges}}

# {{crate}}

{{readme}}

License: {{license}}